package blueeyes.io.iteratee

import Input._

sealed trait Iteratee[E, A] { self =>
  import Iteratee._
  
  /** Applies the iteratee to the source material, given an implicit enumerator factory.
   * {{{
   * countLines(file).end
   * }}}
   */
  def apply[X](source: X)(implicit factory: X => Enumerator[E]): Iteratee[E, A] = factory(source).apply(this)
  
  def map[AA](f: A => AA): Iteratee[E, AA]
  
  def flatMap[AA](f: A => Iteratee[E, AA]): Iteratee[E, AA]
  
  def then[AA](iter: => Iteratee[E, AA]) = flatMap(_ => iter)
  
  def withRemainder(remainder: Input[E]): Iteratee[E, A]
  
  final def >>= [AA](f: A => Iteratee[E, AA]): Iteratee[E, AA] = flatMap[AA](f)
  
  final def >> [AA](iter: Iteratee[E, AA]): Iteratee[E, AA] = then(iter)
  
  /** Feeds the iteratee EOF to produce either an error, or a value.
   */
  def end: Either[IterateeError, A] = {
    def runNext(iteratee: Iteratee[E, A]): Either[IterateeError, A] = {
      iteratee match {
        case Done(v, _)  => Right(v)
        case Next(_)     => error("Diverging iteratee!")
        case Error(e, _) => Left(e)
      }
    }
    
    this match {
      case Done(v, _)   => Right(v)
      case Next(n)      => runNext(n(Eof))
      case Error(e, _)  => Left(e)
    }
  }
}

trait CoreIteratees {
  sealed case class Done[E, A](value: A, remainder: Input[E]) extends Iteratee[E, A] {
    def flatMap[AA](f: A => Iteratee[E, AA]): Iteratee[E, AA] = f(value).withRemainder(remainder)
    
    def map[AA](f: A => AA): Done[E, AA] = Done(f(value), remainder)
    
    def withRemainder(remainder: Input[E]): Done[E, A] = Done(value, remainder)
  }
  
  sealed case class Next[E, A](next: Input[E] => Iteratee[E, A]) extends Iteratee[E, A] {
    def flatMap[AA](f: A => Iteratee[E, AA]): Next[E, AA] = Next((i: Input[E]) =>
      next(i).flatMap(f)
    )
    
    def map[AA](f: A => AA): Next[E, AA] = Next((i: Input[E]) =>
      next(i).map(f)
    )
    
    def withRemainder(remainder: Input[E]): Next[E, A] = Next((i: Input[E]) =>
      next(i).withRemainder(remainder)
    )
  }
  
  sealed case class Error[E, A](error: IterateeError, next: Input[E] => Iteratee[E, A]) extends Iteratee[E, A] {
    def flatMap[AA](f: A => Iteratee[E, AA]): Error[E, AA] = Error(error, (i: Input[E]) =>
      next(i).flatMap(f)
    )
    
    def map[AA](f: A => AA): Error[E, AA] = Error(error, (i: Input[E]) =>
      next(i).map(f)
    )
    
    def withRemainder(remainder: Input[E]): Error[E, A] = Error(error, (i: Input[E]) =>
      next(i).withRemainder(remainder)
    )
  }
  
  final def returnValue[E, A](a: => A): Iteratee[E, A] = Done(a, Empty)
  
  final def returnValue[E, A](a: => A, remainder: Input[E]): Iteratee[E, A] = Done(a, remainder)
  
  final def continueWith[E, A](next: Input[E] => Iteratee[E, A]): Iteratee[E, A] = Next(next)
  
  final def throwErr[E, A](errorValue: IterateeError, nextValue: Input[E] => Iteratee[E, A]): Iteratee[E, A] = Error(errorValue, nextValue)
  
  final def throwFatalErr[E, A](error: IterateeError): Iteratee[E, A] = throwErr[E, A](error, (i: Input[E]) => throwFatalErr(error))
  
  final def remainder[E](remainder: Input[E]): Iteratee[E, Unit] = Done((), remainder)
  
  final def skipToEnd[E]: Iteratee[E, Unit] = continueWith {
    case Eof => returnValue(())
    
    case _ => skipToEnd[E]
  }
  
  /**
   * {{{
   * 
   * }}}
   */
  final def join[E, A](outer: Iteratee[E, Iteratee[E, A]]): Iteratee[E, A] = {
    outer match {
      case Done(a, r) => a
      
      case Next(n) => continueWith((input: Input[E]) =>
        join(n(input))
      )
      
      case Error(e, n) => throwErr(e, (input: Input[E]) => join(n(input)))
    }
  }
  
  final def joinI[E, EE, A](outer: Iteratee[E, Iteratee[EE, A]]): Iteratee[E, A] = {
    outer match {
      case Done(a, r) => a.end match {
        case Left(e)  => throwFatalErr[E, A](e)
        case Right(a) => returnValue[E, A](a)
      }
      
      case Next(n) => continueWith[E, A]((input: Input[E]) =>
        joinI(n(input))
      )
      
      case Error(e, n) => throwErr(e, (input: Input[E]) => joinI(n(input)))
    }
  }
}
object CoreIteratees extends CoreIteratees

object Iteratee extends CoreIteratees {
  implicit def applyForListLikeIteratees[E[_], T, A](iteratee: Iteratee[E[T], A]) = new {
    def apply[X](x: X)(implicit factory: X => Enumerator[E[T]]): Iteratee[E[T], A] = iteratee.apply(x)(factory)
  }
}

import Iteratee._

/** Defines a collection of iteratees that work for any kind of chunk, and 
 * which operate at the chunk level. For all input types whose chunk type
 * is itself a collection of elements, these chunks iteratees are not very 
 * convenient because they are so course-grained. See VectorLikeIteratees.
 */
trait ChunkIteratees extends CoreIteratees {
  def length[E]: Iteratee[E, Int] = {
    def length0(acc: Int): Iteratee[E, Int] = continueWith { 
      case Empty    => length0(acc)
      case Chunk(c) => continueWith((i: Input[E]) => length0(acc + 1))
      case Eof      => returnValue(acc)
    }
  
    length0(0)
  }

  def peek[E]: Iteratee[E, Option[E]] = continueWith {
    case Empty    => peek[E]
    case Chunk(c) => returnValue(Some(c), Chunk(c))
    case Eof      => returnValue(None)
  }

  def head[E]: Iteratee[E, Option[E]] = continueWith {
    case Empty    => head[E]
    case Chunk(c) => returnValue(Some(c))
    case Eof      => returnValue(None)
  }
  
  def takeToList[E](n: Int): Iteratee[E, List[E]] = {
    def take0(count: Int, acc: List[E]): Iteratee[E, List[E]] = {
      if (count <= 0) returnValue(acc)
      else continueWith { 
        case Empty    => take0(count, acc)
        case Chunk(c) => take0(count - 1, c :: acc)
        case Eof      => returnValue(acc)
      }
    }
    
    take0(n, Nil).map(_.reverse)
  }
  
  def takeWhileToList[E](f: E => Boolean): Iteratee[E, List[E]] = {
    def takeWhile0(acc: List[E]): Iteratee[E, List[E]] = continueWith { 
      case Empty    => takeWhile0(acc)
      case Chunk(c) => if (f(c)) takeWhile0(c :: acc) else returnValue(acc, Chunk(c))
      case Eof      => returnValue(acc)
    }
    
    takeWhile0(Nil).map(_.reverse)
  }
  
  def takeAllToList[E]: Iteratee[E, List[E]] = takeWhileToList(e => true)

  def drop[E](n: Int): Iteratee[E, Unit] = {
    if (n <= 0) returnValue(())
    else continueWith { 
      case Empty     => returnValue(())
      case Chunk(c)  => drop[E](n - 1)
      case Eof       => returnValue(())
    }
  }

  def dropWhile[E](f: E => Boolean): Iteratee[E, Unit] = {
    continueWith { 
      case Empty    => returnValue(())
      case Chunk(c) => if (f(c)) dropWhile[E](f) else returnValue((), Chunk(c))
      case Eof      => returnValue(())
    }
  }
}
object ChunkIteratees extends ChunkIteratees

/** Defines a collection of iteratees that work for chunks that are 
 * vector-like -- for example, byte buffers or strings.
 */
trait VectorLikeIteratees extends CoreIteratees {
  def length[E[_], T](implicit vl: VectorLike[E]): Iteratee[E[T], Int] = {
    def length0(acc: Int): Iteratee[E[T], Int] = continueWith {
      case Empty    => length0(acc)
      case Chunk(c) => continueWith((i: Input[E[T]]) => length0(acc + vl.len(c)))
      case Eof      => returnValue(acc)
    }
  
    length0(0)
  }
  
  def peek[E[_], T](implicit vl: VectorLike[E]): Iteratee[E[T], Option[T]] = continueWith {
    case Empty    => peek[E, T]
    case Chunk(c) => vl.headOption(c).map(h => returnValue[E[T], Option[T]](Some(h), Chunk(c))).getOrElse(peek[E, T])
    case Eof      => returnValue(None)
  }
  
  def head[E[_], T](implicit vl: VectorLike[E]): Iteratee[E[T], Option[T]] = continueWith { 
    case Empty    => head[E, T]
    case Chunk(c) => vl.headOption(c).map(h => returnValue[E[T], Option[T]](Some(h), Chunk(vl.tail(c)))).getOrElse(head[E, T])
    case Eof      => returnValue(None)
  }
  
  def takeToList[E[_], T](n: Int)(implicit vl: VectorLike[E]): Iteratee[E[T], List[T]] = {
    def take0(count: Int, acc: List[T]): Iteratee[E[T], List[T]] = {
      if (count <= 0) returnValue(acc)
      else continueWith { 
        case Empty    => take0(count, acc)
        
        case Chunk(c) => {
          val taken = vl.take(c, count)
          val takenLen = vl.len(taken)
          
          val remainder = count - takenLen
          
          val nextAcc = vl.foldLeft(taken, acc) { (acc, t) => t :: acc }
          
          if (remainder > 0) take0(remainder, nextAcc)
          else returnValue(nextAcc, Chunk(vl.drop(c, takenLen)))
        }
        
        case Eof => returnValue(acc)
      }
    }
    
    take0(n, Nil).map(_.reverse)
  }
  
  def takeWhileToList[E[_], T](f: T => Boolean)(implicit vl: VectorLike[E]): Iteratee[E[T], List[T]] = {
    def take0(acc: List[T]): Iteratee[E[T], List[T]] = continueWith {
      case Empty => take0(acc)
      
      case Chunk(c) => {
        val chunkLen = vl.len(c)
        
        val taken    = vl.takeWhile(c, f)
        val takenLen = vl.len(taken)
        
        val mightBeMore = takenLen == chunkLen
        
        val nextAcc = vl.foldLeft(taken, acc) { (acc, t) => t :: acc }
        
        if (mightBeMore) take0(nextAcc)
        else returnValue(nextAcc, Chunk(vl.drop(c, takenLen)))
      }
      
      case Eof => returnValue(acc)
    }
    
    take0(Nil).map(_.reverse)
  }
  
  def takeAllToList[E[_], T](implicit vl: VectorLike[E]): Iteratee[E[T], List[T]] = takeWhileToList(e => true)
  
  def drop[E[_], T](n: Int)(implicit vl: VectorLike[E]): Iteratee[E[T], Unit] = {
    if (n <= 0) returnValue(())
    else continueWith { 
      case Empty => returnValue(())
      
      case Chunk(c) => {
          val chunkLen = vl.len(c)
          
          val rest = vl.drop(c, n)
          val numDropped = chunkLen - vl.len(rest)
          
          val remainder = n - numDropped
          
          if (remainder > 0) drop(remainder)
          else returnValue((), Chunk(rest))
        }
        
      case Eof => returnValue(())
    }
  }
  
  def dropWhile[E[_], T](f: T => Boolean)(implicit vl: VectorLike[E]): Iteratee[E[T], Unit] = continueWith {
    case Empty => returnValue(())
      
    case Chunk(c) => {
      val chunkLen = vl.len(c)
      
      val rest = vl.dropWhile(c, f)
      val numDropped = chunkLen - vl.len(rest)
      
      val mightBeMore = numDropped == chunkLen
      
      if (mightBeMore) dropWhile(f)
      else returnValue((), Chunk(rest))
    }
    
    case Eof => returnValue(())
  }
}
object VectorLikeIteratees extends VectorLikeIteratees

